# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2021-2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
The fparser2 symbol-table module. Defines various classes as well as
the single, global SYMBOL_TABLES instance. The latter is a container
for all of the top-level scoping units encountered during parsing.

"""

from collections import namedtuple


class SymbolTableError(Exception):
    """Base class exception for symbol-table related errors."""


class SymbolTables:
    """
    Class encapsulating functionality for the global symbol-tables object.
    This is a container for all symbol tables constructed while parsing
    code. All names are converted to lower case (since Fortran is not
    case sensitive).

    """

    def __init__(self):
        self._symbol_tables = {}
        # The symbol table of the current scope
        self._current_scope = None
        # Whether or not we enable consistency checks in the symbol tables
        # that are created.
        self._enable_checks = False

    def __str__(self):
        result = (
            f"SymbolTables: {len(self._symbol_tables)} tables\n"
            "========================\n"
        )
        return result + "\n".join(sorted(self._symbol_tables.keys()))

    def enable_checks(self, value):
        """
        Sets whether or not to enable consistency checks in every symbol
        table that is created during a parse.

        :param bool value: whether or not checks are enabled.

        """
        self._enable_checks = value

    def clear(self):
        """
        Deletes any stored SymbolTables.

        """
        self._symbol_tables = {}
        self._current_scope = None

    def add(self, name, node=None):
        """
        Add a new symbol table with the supplied name. The name will be
        converted to lower case if necessary.

        :param str name: the name for the new table.
        :param node: the node in the parse tree associated with this table.
        :type node: Optional[:py:class:`fparser.two.utils.Base`]

        :returns: the new symbol table.
        :rtype: :py:class:`fparser.two.symbol_table.SymbolTable`

        :raises SymbolTableError: if there is already an entry with the
                                  supplied name.
        """
        lower_name = name.lower()
        if lower_name in self._symbol_tables:
            raise SymbolTableError(
                f"The table of top-level (un-nested) symbol tables already "
                f"contains an entry for '{lower_name}'"
            )
        table = SymbolTable(lower_name, checking_enabled=self._enable_checks, node=node)
        self._symbol_tables[lower_name] = table
        return table

    def lookup(self, name):
        """
        Find the named symbol table and return it.

        :param str name: the name of the required symbol table (not case \
                         sensitive).

        :returns: the named symbol table.
        :rtype: :py:class:`fparser.two.symbol_table.SymbolTable`

        """
        return self._symbol_tables[name.lower()]

    @property
    def current_scope(self):
        """
        :returns: the symbol table for the current scoping unit or None.
        :rtype: :py:class:`fparser.two.symbol_table.SymbolTable` or NoneType
        """
        return self._current_scope

    def enter_scope(self, name, node=None):
        """
        Called when the parser enters a new scoping region (i.e. when it
        encounters one of the classes listed in `_scoping_unit_classes`).
        Sets the 'current scope' to be the symbol table with the supplied name.
        If we are not currently within a tree of scoping regions then a
        new entry is created in the internal dict of symbol tables. If there
        is an existing tree then a new table is created and added to the
        bottom.

        :param str name: name of the scoping region.
        :param node: the node of the parse tree associated with this region.
        :type node: Optional[:py:class:`fparser.two.utils.Base`]
        """
        lname = name.lower()

        if not self._current_scope:
            # We're not already inside a nested scope.
            try:
                table = self.lookup(lname)
            except KeyError:
                # Create a new, top-level symbol table with the supplied name.
                table = self.add(lname, node=node)
        else:
            # We are already inside a scoping region so create a new table
            # and setup its parent/child connections.
            table = SymbolTable(
                lname,
                parent=self._current_scope,
                checking_enabled=self._enable_checks,
                node=node,
            )
            self._current_scope.add_child(table)

        # Finally, make this new table the current scope
        self._current_scope = table

    def exit_scope(self):
        """
        Marks the end of the processing of the current scoping unit. Since
        we are exiting the current scoping region, the new 'current scoping
        region' will be its parent.

        :raises SymbolTableError: if there is no current scope from which to \
                                  exit.
        """
        if not self._current_scope:
            raise SymbolTableError("exit_scope() called but no current scope exists.")
        self._current_scope = self._current_scope.parent

    def remove(self, name):
        """
        Removes the named symbol table and any descendants it may have.
        When searching for the named table, the current scope takes priority
        followed by the list of top-level symbol tables.

        :param str name: the name of the symbol table to remove (not case \
                         sensitive).
        :raises SymbolTableError: if the named symbol table is not in the \
            current scope or in the list of top-level symbol tables.

        """
        lname = name.lower()
        if self._current_scope:
            try:
                self._current_scope.del_child(lname)
                # We succeeded in removing it from the current scope so we
                # are done.
                return
            except KeyError:
                pass

        if lname not in self._symbol_tables:
            msg = f"Failed to find a table named '{name}' in "
            if self._current_scope:
                msg += (
                    f"either the current scope (which contains "
                    f"{[child.name for child in self._current_scope.children]}) or "
                )
            msg += (
                f"the list of top-level symbol tables "
                f"({list(self._symbol_tables.keys())})."
            )
            raise SymbolTableError(msg)

        # Check that we're not currently somewhere inside the scope of the
        # named table.
        top_table = self._symbol_tables[lname]
        if self._current_scope:
            if self._current_scope.root is top_table:
                raise SymbolTableError(
                    f"Cannot remove top-level symbol table '{name}' because the "
                    f"current scope '{self._current_scope.name}' has it as an ancestor."
                )

        del self._symbol_tables[lname]


class ModuleUse:
    """
    Class capturing information on all USE statements referring to a given
    Fortran module.

    A USE statement can rename an imported symbol so as to avoid a clash in
    the local scope, e.g. `USE my_mod, alocal => amod` where `amod` is the
    name of the symbol declared in `my_mod`. This renaming can also occur
    inside an Only_List, e.g. `USE my_mod, only: alocal => amod`.

    :param str name: the name of the module.
    :param only_list: list of 2-tuples giving the (local-name, module-name) \
        of symbols that appear in an Only_List. If a symbol is not re-named \
        then module-name can be None.
    :type only_list: Optional[List[Tuple[str, str | NoneType]]]
    :param rename_list: list of 2-tuples given the (local-name, module-name) \
        of symbols that appear in a Rename_List.
    :type rename_list: Optional[List[Tuple[str, str]]]

    :raises TypeError: if any of the supplied parameters are of the wrong type.

    """

    def __init__(self, name, only_list=None, rename_list=None):
        if not isinstance(name, str):
            raise TypeError(
                f"The name of the module must be a str but got "
                f"'{type(name).__name__}'"
            )
        self._validate_tuple_list("only", only_list)
        self._validate_tuple_list("rename", rename_list)

        if only_list and not all(
            isinstance(item[0], str) and (item[1] is None or isinstance(item[1], str))
            for item in only_list
        ):
            raise TypeError(
                f"If present, the only_list must be a list of "
                f"2-tuples of (str, str | NoneType) but got: {only_list}"
            )

        if rename_list and not all(
            isinstance(item[0], str) and isinstance(item[1], str)
            for item in rename_list
        ):
            raise TypeError(
                f"If present, the rename_list must be a list of "
                f"2-tuples of (str, str) but got: {rename_list}"
            )

        self._name = name.lower()

        # dict of Symbols known to be accessed in this module.
        self._symbols = {}
        # Mapping from local symbol name in current scope to actual, declared
        # name in the module from which it is imported.
        self._local_to_module_map = {}

        if only_list is not None:
            self._store_symbols(only_list)
            self._wildcard_import = False
            self._only_set = set(local_name.lower() for local_name, _ in only_list)
        else:
            self._only_set = None
            self._wildcard_import = True

        if rename_list:
            self._store_symbols(rename_list)
            self._rename_set = set(local_name.lower() for local_name, _ in rename_list)
        else:
            self._rename_set = None

    @staticmethod
    def _validate_tuple_list(name, tlist):
        """
        Validate the supplied list of tuples.

        :param str name: the name of the list being validated.
        :param tlist: the list of tuples to validate.
        :type tlist: Optional[List[Tuple[str, str | NoneType]]]

        :raises TypeError: if the supplied list is of the wrong type.

        """
        if not tlist:
            # None or an empty list is fine.
            return

        if not isinstance(tlist, list):
            raise TypeError(
                f"If present, the {name}_list must be a list but "
                f"got '{type(tlist).__name__}'"
            )
        if not all(isinstance(item, tuple) and len(item) == 2 for item in tlist):
            raise TypeError(
                f"If present, the {name}_list must be a list of "
                f"2-tuples but got: {tlist}"
            )

    def _store_symbols(self, rename_list):
        """
        Utility that updates the local list of symbols and renaming map for
        the given list of local-name, module-name tuples. If a symbol is
        not renamed then module-name can be None.

        :param rename_list: list of local-name, module-name pairs.
        :type rename_list: List[Tuple[str, str | NoneType]]

        """
        for local_name, orig_name in rename_list:
            lname = local_name.lower()
            oname = orig_name.lower() if orig_name else None
            if lname not in self._symbols:
                self._symbols[lname] = SymbolTable.Symbol(lname, "unknown")
            if oname:
                self._local_to_module_map[lname] = oname

    def update(self, other):
        """
        Update the current information with the information on the USE held
        in 'other'.

        :param other: the other Use instance from which to update this one.
        :type other: :py:class:`fparser.two.symbol_table.Use`

        :raises TypeError: if 'other' is of the wrong type.
        :raises ValueError: if 'other' is for a module with a different \
                            name to the current one.

        """
        if not isinstance(other, ModuleUse):
            raise TypeError(
                f"update() must be supplied with an instance of "
                f"ModuleUse but got '{type(other).__name__}'"
            )

        if self.name != other.name:
            raise ValueError(
                f"The ModuleUse supplied to update() is for module "
                f"'{other.name}' but this ModuleUse is for module "
                f"'{self.name}'"
            )

        # Take the union of the only and rename lists and update the list
        # of symbols.
        if other.only_list:
            for local_name in other.only_list:
                if local_name not in self._symbols:
                    self._symbols[local_name] = SymbolTable.Symbol(
                        local_name, "unknown"
                    )
            # pylint: disable=protected-access
            if self._only_set is None:
                self._only_set = other._only_set
            else:
                self._only_set = self._only_set.union(other._only_set)
            # pylint: enable=protected-access

        if other.rename_list:
            for local_name in other.rename_list:
                if local_name not in self._symbols:
                    self._symbols[local_name] = SymbolTable.Symbol(
                        local_name, "unknown"
                    )
            # pylint: disable=protected-access
            if self._rename_set is None:
                self._rename_set = other._rename_set
            else:
                self._rename_set = self._rename_set.union(other._rename_set)
            # pylint: enable=protected-access

        # pylint: disable=protected-access
        self._local_to_module_map.update(other._local_to_module_map)
        # pylint: enable=protected-access

        self._wildcard_import = self._wildcard_import or other.wildcard_import

    @property
    def name(self):
        """
        :returns: the name of the Fortran module.
        :rtype: str
        """
        return self._name

    @property
    def symbol_names(self):
        """
        :returns: the names of all symbols associated with USE(s) of this
                  module.
        :rtype: List[str]
        """
        return list(self._symbols.keys())

    def lookup(self, name):
        """
        :returns: the symbol with the supplied name imported from this module (if any).
        :rtype: :py:class:`fparser.two.symbol_table.SymbolTable.Symbol`

        :raises KeyError: if no symbol with the supplied name is imported from
                          this module into the current scope.
        """
        return self._symbols[name.lower()]

    @property
    def only_list(self):
        """
        :returns: the local names that appear in an Only_List or None if there
                  is no such list.
        :rtype: Optional[List[str]]
        """
        if self._only_set is None:
            return None
        return list(self._only_set)

    @property
    def rename_list(self):
        """
        :returns: the local names that appear in a Rename_List or None if there
                  is no such list.
        :rtype: Optional[List[str]]
        """
        if self._rename_set is None:
            return None
        return list(self._rename_set)

    @property
    def wildcard_import(self):
        """
        :returns: whether there is a wildcard import from this module.
        :rtype: bool
        """
        return self._wildcard_import

    def get_declared_name(self, name):
        """
        :returns: the name of the supplied symbol as declared in the module.
        :rtype: str
        """
        return self._local_to_module_map.get(name, name)


class SymbolTable:
    """
    Class implementing a single symbol table.

    Since this functionality is not yet fully mature, checks that new symbols
    don't clash with existing symbols are disabled by default.
    Once #201 is complete it is planned to switch this so that the checks
    are instead enabled by default.

    :param str name: the name of this scope. Will be the name of the
        associated module or routine.
    :param parent: the symbol table within which this one is nested (if any).
    :type parent: :py:class:`fparser.two.symbol_table.SymbolTable.Symbol`
    :param bool checking_enabled: whether or not validity checks are
        performed for symbols added to the table.
    :param node: the node in the parse tree associated with this table.
    :type node: Optional[:py:class:`fparser.two.utils.Base`]

    :raises TypeError: if the supplied node is of the wrong type.
    """

    # TODO #201 add support for other symbol properties (kind, shape
    # and visibility). We may need a distinct Symbol class so as to provide
    # type checking for the various properties.
    Symbol = namedtuple("Symbol", "name primitive_type")

    def __init__(self, name, parent=None, checking_enabled=False, node=None):
        self._name = name.lower()
        # Symbols defined in this scope that represent data.
        self._data_symbols = {}
        # dict of ModuleUse objects (indexed by module name) representing
        # modules imported into this scope.
        self._modules = {}
        # Reference to a SymbolTable that contains this one (if any). Actual
        # value (if any) is set via setter method.
        self._parent = None
        self.parent = parent
        # The node in the parse tree with which this table is associated (if any).
        from fparser.two.utils import Base

        if node and not isinstance(node, Base):
            raise TypeError(
                f"The 'node' argument to the SymbolTable constructor must be a "
                f"valid parse tree node (instance of utils.Base) but got "
                f"'{type(node).__name__}'"
            )
        self._node = node
        # Whether or not to perform validity checks when symbols are added.
        self._checking_enabled = checking_enabled
        # Symbol tables nested within this one.
        self._children = []

    def __str__(self):
        header = "===========\n"
        symbols = "Symbols:\n"
        if self._data_symbols:
            symbols += "\n".join(list(self._data_symbols.keys())) + "\n"
        uses = "Used modules:\n"
        if self._modules:
            uses += "\n".join(list(self._modules.keys())) + "\n"
        return f"{header}Symbol Table '{self._name}'\n{symbols}{uses}{header}"

    def add_data_symbol(self, name, primitive_type):
        """
        Creates a new Symbol with the specified properties and adds it to
        the symbol table. The supplied name is converted to lower case.

        TODO #201 add support for other symbol properties (kind, shape
        and visibility).

        :param str name: the name of the symbol.
        :param str primitive_type: the primitive type of the symbol.

        :raises TypeError: if any of the supplied parameters are of the \
                           wrong type.
        :raises SymbolTableError: if the symbol table already contains an
                                  entry with the supplied name.
        """
        if not isinstance(name, str):
            raise TypeError(
                f"The name of the symbol must be a str but got "
                f"'{type(name).__name__}'"
            )
        # TODO #201 use an enumeration for the primitive type
        if not isinstance(primitive_type, str):
            raise TypeError(
                f"The primitive type of the symbol must be specified as a str "
                f"but got '{type(primitive_type).__name__}'"
            )
        lname = name.lower()

        if self._checking_enabled:
            if lname in self._data_symbols:
                raise SymbolTableError(
                    f"Symbol table already contains a symbol for"
                    f" a variable with name '{name}'"
                )
            if lname in self._modules:
                raise SymbolTableError(
                    f"Symbol table already contains a use of a "
                    f"module with name '{name}'"
                )
            for mod_name, mod in self._modules.items():
                if mod.symbol_names and lname in mod.symbol_names:
                    raise SymbolTableError(
                        f"Symbol table already contains a use of a symbol "
                        f"named '{name}' from module '{mod_name}'"
                    )

        self._data_symbols[lname] = SymbolTable.Symbol(lname, primitive_type.lower())

    def add_use_symbols(self, name, only_list=None, rename_list=None):
        """
        Creates an entry in the table for the USE of a module with the supplied
        name. If no `only_list` is supplied then this USE represents a wildcard
        import of all public symbols in the named module. If the USE statement
        has an ONLY clause but without any named symbols then `only_list`
        should be an empty list.

        A USE can also have one or more rename entries *without* an only list.

        :param str name: the name of the module being imported via a USE. Not \
            case sensitive.
        :param only_list: if there is an 'only:' clause on the USE statement \
            then this contains a list of tuples, each holding the local name \
            of the symbol and its name in the module from which it is \
            imported. These names are case insensitive.
        :type only_list: Optional[List[Tuple[str, str | NoneType]]]
        :param rename_list: a list of symbols that are renamed from the scope \
            being imported. Each entry is a tuple containing the name in the \
            local scope and the corresponding name in the module from which it\
            is imported. These names are case insensitive.
        :type rename_list: Optional[List[Tuple[str, str]]]

        """
        use = ModuleUse(name, only_list, rename_list)

        if use.name in self._modules:
            # The same module can appear in more than one use statement
            # in Fortran.
            self._modules[use.name].update(use)
        else:
            self._modules[use.name] = use

    def lookup(self, name):
        """
        Lookup the symbol with the supplied name.

        :param str name: the name of the symbol to lookup (not case sensitive).

        :returns: the named symbol.
        :rtype: :py:class:`fparser.two.symbol_table.SymbolTable.Symbol`

        :raises KeyError: if the named symbol cannot be found in this or any \
                          parent scope.
        """
        # Fortran is not case sensitive so convert input to lowercase.
        lname = name.lower()
        if lname in self._data_symbols:
            # Found a match in this table.
            return self._data_symbols[lname]
        for module in self._modules.values():
            try:
                # Look to see whether the symbol is imported into this table.
                return module.lookup(lname)
            except KeyError:
                pass
        # No match in this scope - search in parent scope (if any). This will
        # recurse upwards through parent tables as necessary.
        if self.parent:
            return self.parent.lookup(lname)
        raise KeyError(f"Failed to find symbol named '{lname}'")

    @property
    def name(self):
        """
        :returns: the name of this symbol table (scoping region).
        :rtype: str
        """
        return self._name

    @property
    def parent(self):
        """
        :returns: the parent symbol table (scoping region) that contains \
                  this one (if any).
        :rtype: :py:class:`fparser.two.symbol_table.SymbolTable` or NoneType
        """
        return self._parent

    @parent.setter
    def parent(self, value):
        """
        Set the parent scope for this symbol table.

        :param value: the parent symbol table.
        :type value: :py:class:`fparser.two.symbol_table.SymbolTable`

        :raises TypeError: if the supplied value is not None or a SymbolTable.

        """
        if value is not None and not isinstance(value, SymbolTable):
            raise TypeError(
                f"Unless it is None, the parent of a SymbolTable must also be "
                f"a SymbolTable but got '{type(value).__name__}'"
            )
        self._parent = value

    @property
    def node(self):
        """
        :returns: the scoping node (in the parse tree) asssociated with this SymbolTable.
        :rtype: :py:class:`fparser.two.utils.Base`
        """
        return self._node

    def add_child(self, child):
        """
        Adds a child symbol table (scoping region nested within this one).

        :param child: the nested symbol table.
        :type child: :py:class:`fparser.two.symbol_table.SymbolTable`

        :raises TypeError: if the supplied child is not a SymbolTable.

        """
        if not isinstance(child, SymbolTable):
            raise TypeError(
                f"Expected a SymbolTable instance but got '{type(child).__name__}'"
            )
        self._children.append(child)

    def del_child(self, name):
        """
        Removes the named symbol table.

        :param str name: the name of the child symbol table to delete (not \
                         case sensitive).

        :raises KeyError: if the named table is not a child of this one.

        """
        lname = name.lower()
        for child in self._children:
            if child.name == lname:
                self._children.remove(child)
                break
        else:
            raise KeyError(
                f"Symbol table '{self.name}' does not contain a table named '{name}'"
            )

    @property
    def children(self):
        """
        :returns: the child (nested) symbol tables, if any.
        :rtype: list of :py:class:`fparser.two.symbol_table.SymbolTable`
        """
        return self._children

    @property
    def root(self):
        """
        :returns: the top-level symbol table that contains the current \
                  scoping region (symbol table).
        :rtype: :py:class:`fparser.two.symbol_table.SymbolTable`

        """
        current = self
        while current.parent:
            current = current.parent
        return current

    @property
    def wildcard_imports(self):
        """
        :returns: names of all modules with wildcard imports into this scope or an
                  empty list if there are none.
        :rtype: List[Optional[str]]
        """
        mod_names = set()
        for mod_name, mod in self._modules.items():
            if mod.wildcard_import:
                mod_names.add(mod_name)
        if self.parent:
            # Any wildcard imports in a parent scope will affect this scoping
            # region so carry on up. Note that if the root scoping region in
            # the current file is a SUBMODULE then we will be missing whatever
            # is brought into scope in the parent MODULE (since that will typically
            # be in a separate source file).
            mod_names.update(self.parent.wildcard_imports)

        return sorted(list(mod_names))

    @property
    def all_symbols_resolved(self):
        """
        :returns: whether all symbols in this scope have been resolved. i.e. if
            there are any wildcard imports or this table is within a submodule
            then there could be symbols we don't have definitions for.
        :rtype: bool
        """
        # wildcard_imports checks all parent scopes.
        if self.wildcard_imports:
            return False

        # pylint: disable=import-outside-toplevel
        from fparser.two.Fortran2008 import Submodule_Stmt

        cursor = self
        while cursor:
            if isinstance(cursor.node, Submodule_Stmt):
                return False
            cursor = cursor.parent
        return True


#: The single, global container for all symbol tables constructed while
#: parsing.
SYMBOL_TABLES = SymbolTables()


__all__ = ["SymbolTableError", "SymbolTables", "SymbolTable", "SYMBOL_TABLES"]



# Modified work Copyright (c) 2017 Science and Technology Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
from fparser import api


def test_default_private():
    src = """\
module mod1
private
integer :: i
contains
subroutine s1
end subroutine
end module mod1
"""
    mod1 = api.parse(src, isfree=True, isstrict=False).content[0]
    assert mod1.get_provides() == {}, repr(mod1.get_provides())
    assert list(mod1.a.variables.keys()) == ["i"]
    assert list(mod1.a.module_subprogram.keys()) == ["s1"]


def test_access_spec():
    src = """\
module mod1
private
integer, public :: i
integer :: j, k
public :: j, s1
contains
subroutine s1
end subroutine
subroutine s2
end subroutine
end module mod1
"""
    mod1 = api.parse(src, isfree=True, isstrict=False).content[0]
    assert sorted(mod1.get_provides().keys()) == sorted(["i", "j", "s1"])
    assert sorted(mod1.a.variables.keys()) == sorted(["i", "j", "k"])
    assert sorted(mod1.a.module_subprogram.keys()) == sorted(["s2", "s1"])



# This Python file uses the following encoding: utf-8
# Copyright (c) 2022 Science and Technology Facilities Council.
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""pytest module for the Fortran2003 Allocate Statement - R263."""

import pytest
from fparser.two.utils import NoMatchError
from fparser.two.Fortran2003 import Allocate_Stmt, Alloc_Opt, Alloc_Opt_List


@pytest.mark.usefixtures("f2003_create")
def test_allocate_stmt():
    """Tests for the allocate statement: R623."""
    tcls = Allocate_Stmt
    assert tcls.alloc_opt_list() == Alloc_Opt_List

    obj = tcls("allocate(a,b)")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "ALLOCATE(a, b)"

    obj = tcls("allocate(real::a)")
    assert str(obj) == "ALLOCATE(REAL::a)"

    obj = tcls("allocate(real(kind=8)::a, stat=b, source=c//d)")
    assert str(obj) == "ALLOCATE(REAL(KIND = 8)::a, STAT = b, SOURCE = c // d)"


@pytest.mark.usefixtures("f2003_create")
def test_allocate_no_match():
    """Tests that the expected NoMatchError is raised if there are problems."""
    tcls = Allocate_Stmt
    # Missing parenthesis.
    with pytest.raises(NoMatchError) as err:
        tcls("allocate(var(3)")
    assert "allocate(var(3)" in str(err.value)
    with pytest.raises(NoMatchError) as err:
        tcls("allocate var(3))")
    assert "allocate var(3))" in str(err.value)
    # Misspelt key word.
    with pytest.raises(NoMatchError) as err:
        tcls("allocte(var(3))")
    assert "allocte(var(3))" in str(err.value)
    # No arguments.
    with pytest.raises(NoMatchError) as err:
        tcls("allocate()")
    assert "allocate()" in str(err.value)
    # Missing positional argument.
    with pytest.raises(NoMatchError) as err:
        tcls("allocate(stat=ierr)")
    assert "allocate(stat=ierr)" in str(err.value)


@pytest.mark.usefixtures("f2003_create")
def test_alloc_opt():
    """Tests for the various forms of alloc-opt: R624."""
    tcls = Alloc_Opt
    obj = tcls("stat=a")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "STAT = a"
    assert repr(obj) == "Alloc_Opt('STAT', Name('a'))"
    obj = tcls("errmsg=my_msg")
    assert str(obj) == "ERRMSG = my_msg"
    obj = tcls("source=b")
    assert str(obj) == "SOURCE = b"
    # Check for a failed match - use 'mold' as that's a Fortran2008 addition
    # so should not match here.
    with pytest.raises(NoMatchError):
        tcls("MOLD=b")
    with pytest.raises(NoMatchError):
        tcls("value")



# Copyright (c) 2023 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test intrinsic handling within Fortran2008. At the moment, the only
special consideration (beyond 2003) is that we can't disambiguate a
shadowed intrinsic within a submodule.

"""

import pytest
from fparser.api import get_reader
from fparser.two import Fortran2003, Fortran2008
from fparser.two.utils import walk


@pytest.mark.usefixtures("f2008_create")
def test_intrinsic_in_submodule():
    """Test that a shadowed intrinsic within a submodule is accepted
    (since a submodule shares the scope of its parent module and we
    assume that is bringing the shadowed symbol into scope).

    """
    reader = get_reader("""\
      submodule (foobar) bar

        contains
        subroutine my_sub
          real :: a
          a = dot_product(1,2,3)
        end subroutine my_sub
      end
      """)
    ast = Fortran2008.Submodule(reader)
    assert not walk(ast, Fortran2003.Intrinsic_Function_Reference)



# Copyright (c) 2021 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R1232 : the majority of the tests for this
are still in test_fortran2003.py and need to be moved here TODO #306.

"""

import pytest
from fparser.api import get_reader
from fparser.two.Fortran2003 import Subroutine_Subprogram, Subroutine_Stmt, Name
from fparser.two.symbol_table import SYMBOL_TABLES


def test_sub_stmt_new_symbol_table(f2003_create):
    """
    Test that valid code is parsed correctly and an associated symbol table
    created.

    """
    obj = Subroutine_Subprogram(get_reader("subroutine a\nend subroutine"))
    assert isinstance(obj, Subroutine_Subprogram)
    assert str(obj) == "SUBROUTINE a\nEND SUBROUTINE"
    assert repr(obj) == (
        "Subroutine_Subprogram(Subroutine_Stmt(None, "
        "Name('a'), None, None), End_Subroutine_Stmt('"
        "SUBROUTINE', None))"
    )
    assert "a" in SYMBOL_TABLES._symbol_tables


def test_subroutine_get_name():
    """Test we can get the name of the subroutine"""
    obj = Subroutine_Stmt("subroutine foo")
    assert obj.get_name() == Name("foo")



# Modified work Copyright (c) 2017 Science and Technology Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

from fparser import api


def test_reproduce_issue_private():
    source_str = """
    module m
    contains
    subroutine a
    end subroutine a
    end module m
    """
    tree = api.parse(source_str, isfree=True, isstrict=False)


def test_private_subroutine():
    source_str = """
    module m
    public
    private a
    contains
    subroutine a
    end subroutine a
    subroutine b
    end subroutine b
    end module m
    """
    tree = api.parse(source_str, isfree=True, isstrict=False)
    a = tree.content[0].content[3]
    b = tree.content[0].content[4]

    assert not a.is_public()
    assert a.is_private()

    assert b.is_public()
    assert not b.is_private()


def test_related_issue_type():
    source_str = """
    module m
    type private :: a
    end type a
    type public :: b
    end type b
    type :: c
    end type c
    end module m
    """
    tree = api.parse(source_str, isfree=True, isstrict=False)
    a, b, c = tree.content[0].content[:3]
    assert a.is_private()
    assert not a.is_public()

    assert not b.is_private()
    assert b.is_public()

    assert not c.is_private()
    assert c.is_public()


def test_private_type():
    source_str = """
    module m
    private
    public b
    type :: a
    end type a
    type :: b
    end type b
    type public :: c
    end type c
    end module m
    """
    tree = api.parse(source_str, isfree=True, isstrict=False)
    a, b, c = tree.content[0].content[2:5]
    assert a.is_private()
    assert not a.is_public()
    assert not b.is_private()
    assert b.is_public()
    assert not c.is_private()
    assert c.is_public()


def test_reproduce_issue_len():
    source_str = """
    subroutine foo(a)
    character(lenmax) a
    character(lenmax, kind=4) b
    character(len=lenmax, kind=4) c
    character(kind=4, len=lenmax) d
    character(lenmax, 4) e
    end subroutine foo
    """
    tree = api.parse(source_str, isfree=True, isstrict=False)
    a, b, c, d, e = tree.content[0].content[:5]
    assert a.selector == ("lenmax", "")
    assert b.selector == ("lenmax", "4")
    assert c.selector == ("lenmax", "4")
    assert d.selector == ("lenmax", "4")
    assert e.selector == ("lenmax", "4")



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Executable_Construct rule R213
"""

from fparser.two.Fortran2003 import Executable_Construct as Executable_Construct_2003


class Executable_Construct(Executable_Construct_2003):  # R213
    # pylint: disable=invalid-name
    """
    Fortran 2008 rule R213.

    .. code-block:: fortran

        executable-construct is action-stmt
                             or associate-construct
                             or block-construct
                             or case-construct
                             or critical-construct
                             or do-construct
                             or forall-construct
                             or if-construct
                             or select-type-construct
                             or where-construct

    Associated constraints are:

    "C201 (R208) An execution-part shall not contain an end-function-stmt,
          end-mp-subprogram-stmt, end-program-stmt, or end-subroutine-stmt."

    """

    subclass_names = [
        "Action_Stmt",
        "Associate_Construct",
        "Block_Construct",
        "Case_Construct",
        "Critical_Construct",
        "Do_Construct",
        "Forall_Construct",
        "If_Construct",
        "Select_Type_Construct",
        "Where_Construct",
    ]



# Copyright (c) 2018-2022 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R201 : This file tests the support for zero
or more program-units.

"""

import pytest
from fparser.two.utils import FortranSyntaxError
from fparser.api import get_reader
from fparser.two.Fortran2003 import Program

# Test no content or just white space. This is not officially a
# Fortran rule but fortran compilers tend to accept empty content so
# we follow their lead.


def test_empty_input(f2003_create):
    """Test that empty input or input only containing white space can be
    parsed succesfully

    """
    for code in ["", "   ", "  \n  \n\n"]:
        reader = get_reader(code)
        ast = Program(reader)
        assert str(ast) == ""


@pytest.mark.parametrize("ignore_comments", [True, False])
def test_only_comments(f2003_create, ignore_comments):
    """Test that a file containing only comments can be parsed
    successfully, irrespective of whether or not we are ignoring
    comments.

    """
    code = "! comment1\n! comment2"
    reader = get_reader(code, ignore_comments=ignore_comments)
    ast = Program(reader)
    if ignore_comments:
        assert str(ast) == ""
    else:
        assert code in str(ast)


# Test single program units


@pytest.mark.usefixtures("f2003_create")
def test_single():
    """Test that a single program_unit can be parsed successfully."""
    reader = get_reader("""\
      subroutine test()
      end subroutine
      """)
    ast = Program(reader)
    assert "SUBROUTINE test\n" "END SUBROUTINE" in str(ast)
    # Check that the Name of the subroutine has the correct parent
    assert ast.content[0].content[0].items[1].parent is ast.content[0].content[0]


@pytest.mark.usefixtures("f2003_create")
def test_single_with_end_name():
    """Test that a single program_unit can be parsed successfully when it
    has a name specified on the end clause."""
    reader = get_reader("""\
      subroutine test()
      end subroutine test
      """)
    ast = Program(reader)
    assert "SUBROUTINE test\n" "END SUBROUTINE test" in str(ast)
    # Check parent information has been set-up correctly
    end_sub = ast.content[0].content[-1]
    assert end_sub.items[1].parent is end_sub


@pytest.mark.xfail(
    reason="5 spaces causes the error exception to occur at " "the end of the file"
)
def test_single2(f2003_create):
    """Test that a single program_unit with 5 or more spaces at the start
    of the line reports an error on the correct (first) line

    """
    reader = get_reader("""\
     subroutin test()
     end subroutine

      """)
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy_ = Program(reader)
    assert "at line 1\n>>>      subroutin test()\n" in str(excinfo.value)


@pytest.mark.xfail(
    reason="5 spaces causes the error exception to occur at " "the end of the file"
)
def test_single3(f2003_create):
    """Test that a single program_unit with 5 or more spaces at the start
    of the line reports an error on the correct (second) line

    """
    reader = get_reader("""\
     subroutine test()
     end subroutin

      """)
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy_ = Program(reader)
    assert "at line 2\n>>>      end subroutin\n" in str(excinfo.value)


def test_single_error1(f2003_create):
    """Test that a single program_unit with an error in the initial
    statement raises an appropriate exception

    """
    reader = get_reader("""\
      subroutin test()
      end subroutine
      """)
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy_ = Program(reader)
    assert "at line 1\n>>>      subroutin test()\n" in str(excinfo.value)


def test_single_error2(f2003_create):
    """Test that a single program_unit with an error in the final
    statement raises an appropriate exception

    """
    reader = get_reader("subroutine test()\n\n" "end subroutin\n\n\n")
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy = Program(reader)
    assert "at line 3\n>>>end subroutin\n" in str(excinfo.value)


# Test multiple program units


def test_multiple(f2003_create):
    """Test that multiple program_units can be parsed successfully."""
    reader = get_reader("""\
      subroutine test()
      end subroutine
      subroutine test2()
      end subroutine test2
      """)
    ast = Program(reader)
    assert "SUBROUTINE test\n" "END SUBROUTINE" in str(ast)


@pytest.mark.xfail(
    reason="Having the same name in different program_units "
    "does not raise an exception"
)
def test_multiple_error1(f2003_create):
    """Test that multiple program_units with the same name raise an
    exception

    """
    reader = get_reader("""\
      subroutine test()
      end subroutine
      subroutine test()
      end subroutine
      """)
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy_ = Program(reader)
    assert "XXX" in str(excinfo.value)


def test_multiple_error2(f2003_create):
    """Test that a single program_unit with an error raises an appropriate
    exception

    """
    reader = get_reader("""\
      subroutine 1test()
      end subroutine
      subroutine test()
      end subroutine
      """)
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy_ = Program(reader)
    assert "at line 1\n>>>      subroutine 1test()\n" in str(excinfo.value)


def test_multiple_error3(f2003_create):
    """Test that multiple program_units with an error raises an
    appropriate exception

    """
    reader = get_reader("""\
      subroutine test()
      end subroutine
      subroutine test()
      end subroutin""")
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy_ = Program(reader)
    assert "at line 4\n>>>      end subroutin\n" in str(excinfo.value)


# Test a program unit with a missing program statement


def test_missing_prog(f2003_create):
    """Test that a main program program_unit without a program declaration
    can be parsed successfully. This should not really be a test here,
    but this case is currently treated separately by the match method
    in Program.

    """
    reader = get_reader("""\
      end
      """)
    ast = Program(reader)
    assert "END" in str(ast)
    reader = get_reader(
        """\
      ! This is a program without a program declaration
      end
      """,
        isfree=True,
        ignore_comments=False,
    )
    ast = Program(reader)
    assert "END" in str(ast)


@pytest.mark.xfail(reason="Only the main program is output")
def test_missing_prog_multi(f2003_create):
    """Test that a main program program_unit without a program declaration
    can be parsed successfully when it is not the first program_unit.

    """
    reader = get_reader("""\
      subroutine first
      end
      end
      """)
    ast = Program(reader)
    assert "SUBROUTINE first\n" "END SUBROUTINE" in str(ast)
    assert "END PROGRAM" in str(ast)


# A program should contain only only one main program


@pytest.mark.xfail(reason="Only one main program is allowed in a program")
def test_one_main1(f2003_create):
    """Test that multiple main programs raise an exception."""
    reader = get_reader("""\
      program first
      end
      program second
      end
      """)
    with pytest.raises(FortranSyntaxError) as excinfo:
        dummy_ = Program(reader)
    assert "XXX" in str(excinfo.value)


# Check comments are supported at this level


def test_comment0(f2003_create):
    """Test that a single program_unit without comments can be parsed
    successfully with comment processing switched on.

    """
    reader = get_reader(
        ("subroutine test()\n" "end subroutine\n"), ignore_comments=False
    )
    ast = Program(reader)
    assert ("SUBROUTINE test\n" "END SUBROUTINE") in str(ast)


def test_comment1(f2003_create):
    """Test that a single program_unit can be parsed successfully with
    comments being ignored."""
    reader = get_reader("""\
      ! comment1
      subroutine test()
      end subroutine
      ! comment2
      """)
    ast = Program(reader)
    assert "SUBROUTINE test\n" "END SUBROUTINE" in str(ast)
    assert "! comment" not in str(ast)


def test_comment2(f2003_create):
    """Test that a single program_unit can be parsed successfully with
    comments being included."""
    reader = get_reader(
        ("! comment1\n" "subroutine test()\n" "end subroutine\n" "! comment2\n"),
        ignore_comments=False,
    )
    ast = Program(reader)
    assert ("! comment1\n" "SUBROUTINE test\n" "END SUBROUTINE\n" "! comment2") in str(
        ast
    )


def test_comment3(f2003_create):
    """Test that multiple program_units can be parsed successfully with
    comments being ignored."""
    reader = get_reader(
        """\
      ! comment1
      subroutine test()
      end subroutine
      ! comment2
      module example
      end module
      ! comment3
      """,
        ignore_comments=True,
    )
    ast = Program(reader)
    assert (
        "SUBROUTINE test\n" "END SUBROUTINE\n" "MODULE example\n" "END MODULE"
    ) in str(ast)
    assert "! comment" not in str(ast)


def test_comment4(f2003_create):
    """Test that multiple program_units can be parsed successfully with
    comments being included."""
    reader = get_reader(
        (
            "! comment1\n"
            "subroutine test()\n"
            "end subroutine\n"
            "! comment2\n"
            "module example\n"
            "end module\n"
            "! comment3\n"
        ),
        ignore_comments=False,
    )
    ast = Program(reader)
    assert (
        "! comment1\n"
        "SUBROUTINE test\n"
        "END SUBROUTINE\n"
        "! comment2\n"
        "MODULE example\n"
        "END MODULE\n"
        "! comment3"
    ) in str(ast)


# Check includes are supported at this level


def test_include0(f2003_create):
    """Test that a single program_unit with includes can be parsed
    succesfully.

    """
    reader = get_reader(
        ("include '1'\n" "subroutine test()\n" "end subroutine\n" "include '2'\n")
    )
    ast = Program(reader)
    assert (
        "INCLUDE '1'\n" "SUBROUTINE test\n" "END SUBROUTINE\n" "INCLUDE '2'"
    ) in str(ast)


def test_include1(f2003_create):
    """Test that multiple program_units with includes can be parsed
    successfully.

    """
    reader = get_reader(
        "include '1'\n"
        "subroutine test()\n"
        "end subroutine\n"
        "include '2'\n"
        "module example\n"
        "end module\n"
        "include '3'\n",
        ignore_comments=True,
    )
    ast = Program(reader)
    assert (
        "INCLUDE '1'\n"
        "SUBROUTINE test\n"
        "END SUBROUTINE\n"
        "INCLUDE '2'\n"
        "MODULE example\n"
        "END MODULE\n"
        "INCLUDE '3'"
    ) in str(ast)
    assert "! comment" not in str(ast)


# Check a mix of includes and comments are supported at this level


def test_mix(f2003_create):
    """Test that multiple program_units can be parsed successfully with a
    mix of includes and comments.

    """
    reader = get_reader(
        (
            "include '1'\n"
            "! comment1\n"
            "include '2'\n"
            "subroutine test()\n"
            "end subroutine\n"
            "include '3'\n"
            "include '4'\n"
            "! comment2\n"
            "! comment3\n"
            "module example\n"
            "end module\n"
            "! comment4\n"
            "include '5'\n"
            "! comment5\n"
        ),
        ignore_comments=False,
    )
    ast = Program(reader)
    assert (
        "INCLUDE '1'\n"
        "! comment1\n"
        "INCLUDE '2'\n"
        "SUBROUTINE test\n"
        "END SUBROUTINE\n"
        "INCLUDE '3'\n"
        "INCLUDE '4'\n"
        "! comment2\n"
        "! comment3\n"
        "MODULE example\n"
        "END MODULE\n"
        "! comment4\n"
        "INCLUDE '5'\n"
        "! comment5"
    ) in str(ast)



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Data_Component_Def_Stmt rule R436
"""

from fparser.two.Fortran2003 import (
    Data_Component_Def_Stmt as Data_Component_Def_Stmt_2003,
)
from fparser.two.Fortran2003 import Declaration_Type_Spec, Component_Decl_List
from fparser.two.utils import Type_Declaration_StmtBase


class Data_Component_Def_Stmt(Data_Component_Def_Stmt_2003):  # R436
    """
    Fortran 2008 rule 436.

    .. code-block:: fortran

        data-component-def-stmt is declaration-type-spec [
                 [ , component-attr-spec-list ] :: ] component-decl-list

    The implementation of this rule does not add anything to the Fortran 2003
    variant but reimplements the match method identical to Fortran 2003 as
    otherwise the generated Fortran 2008 variant of `Component_Attr_Spec_List`
    would not be used. Unfortunately, the required `attr_spec_list_cls` can not
    simply be provided as a class property since the relevant class is only
    generated at the end of this file using the `use_names` class property of
    this class.

    Associated constraints are:

    "C439 (R436)  No component-attr-spec shall appear more than once in a given
          component-def-stmt."
    "C440 (R436)  If neither the POINTER nor the ALLOCATABLE attribute is
          specified, the declaration-type-spec in the component-def-stmt shall
          specify an intrinsic type or a previously defined derived type."
    "C441 (R436)  If the POINTER or ALLOCATABLE attribute is specified, each
          component-array-spec shall be a deferred-shape-spec-list."
    "C442 (R436)  If a coarray-spec appears, it shall be a
          deferred-coshape-spec-list and the component shall have the
          ALLOCATABLE attribute."
    "C443 (R436)  If a coarray-spec appears, the component shall not be of type
          C_PTR or C_FUNPTR."
    "C445 (R436)  If neither the POINTER nor the ALLOCATABLE attribute is
          specified, each component-array-spec shall be an
          explicit-shape-spec-list."
    "C447 (R436)  A component shall not have both the ALLOCATABLE and POINTER
          attributes."
    "C448 (R436)  If the CONTIGUOUS attribute is specified, the component shall
          be an array with the POINTER attribute."
    "C457 (R436)  If component-initialization appears, a double-colon separator
          shall appear before the component-decl-list."
    "C458 (R436)  If component-initialization appears, every type parameter and
          array bound of the component shall be a colon or constant expression.
    "C459 (R436)  If => appears in component-initialization, POINTER shall
          appear in the component-attr-spec-list. If = appears in
          component-initialization, neither POINTER nor ALLOCATABLE shall
          appear in the component-attr-spec-list."

    C439-C443, C445, C447-C448, C457-C459 are currently not checked
    - issue #258.

    """

    @staticmethod
    def match(string):
        """Implements the matching of a data component definition statement.

        :param str string: the reader or string to match as a data \
                           component definition statement.

        :return: a 3-tuple containing declaration type specification, \
                 component attribute specification and component declaration \
                 list if there is a match or None if there is no match.
        :rtype: `NoneType` or \
            (:py:class:`fparser.two.Fortran2003.Declaration_Type_Spec`, \
             :py:class:`fparser.two.Fortran2008.Component_Attr_Spec_List`, \
             :py:class:`fparser.two.Fortran2003.Component_Decl_List`)

        """
        # Avoid circular dependencies for generated clas by importing here.
        # pylint: disable=import-outside-toplevel
        from fparser.two.Fortran2008 import Component_Attr_Spec_List

        return Type_Declaration_StmtBase.match(
            Declaration_Type_Spec, Component_Attr_Spec_List, Component_Decl_List, string
        )



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""Test Fortran 2008 rule R817

    nonlabel-do-stmt is [ do-construct-name : ] DO [ loop-control ]

The only difference to F2003 rule R829 is that we force this rule to
use the F2008 version of loop-control

"""

import pytest

from fparser.api import get_reader
from fparser.two.Fortran2008 import Nonlabel_Do_Stmt


@pytest.mark.usefixtures("f2008_create")
def test_concurrent():
    """Test that the Fortran2008 version supports do concurrent."""
    code = "DO CONCURRENT (i = 1 : 20)"
    reader = get_reader(code)
    obj = Nonlabel_Do_Stmt(reader)
    assert isinstance(obj, Nonlabel_Do_Stmt)
    assert str(obj) == code


# The 2008 version of the Nonlabel_Do_Stmt class is only added to make
# sure that that a labelled do concurrent is parsed in f2008. There is
# already a functional tests (called test_functional) in
# test_block_nonlabel_do_construct_r814_2.py which make sure this
# class does its job.



# This Python file uses the following encoding: utf-8
# Copyright (c) 2019-2022 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""pytest module for the Fortran2003 Case Construct - R808.
Does not test all aspects of R808, in particular the conditions C803-7
are not checked - #232."""

import pytest
from fparser.api import get_reader
from fparser.common.readfortran import FortranStringReader
from fparser.two.Fortran2003 import Case_Construct
from fparser.two.utils import FortranSyntaxError


@pytest.mark.usefixtures("f2003_create")
def test_case_construct():
    """Basic test that we parse a Case Construct successfully."""
    tcls = Case_Construct
    obj = tcls(get_reader("""\
select case (n)
case (:-1)
  signum = -1
case (0)
  signum = 0
case (1:)
  signum = 1
case default
  signum = -2
end select
"""))
    assert isinstance(obj, tcls), repr(obj)
    assert (
        str(obj) == "SELECT CASE (n)\nCASE (: - 1)\n  signum = - 1\nCASE (0)\n"
        "  signum = 0\nCASE (1 :)\n  signum = 1\nCASE DEFAULT\n"
        "  signum = - 2\nEND SELECT"
    )


def test_case_construct_name(f2003_create):
    """Basic test that we parse a Case Construct successfully."""
    tcls = Case_Construct
    obj = tcls(get_reader("""\
    name: select case (n)
    case (:-1) name
        signum = -1
    case (0) name
        signum = 0
    case (1:) name
        signum = 1
    case default name
        signum = -2
    end select name
    """))
    assert isinstance(obj, tcls), repr(obj)
    assert (
        str(obj)
        == "name:SELECT CASE (n)\nCASE (: - 1) name\n  signum = - 1\nCASE (0) name\n"
        "  signum = 0\nCASE (1 :) name\n  signum = 1\nCASE DEFAULT name\n"
        "  signum = - 2\nEND SELECT name"
    )


@pytest.mark.usefixtures("f2003_create")
def test_tofortran_non_ascii():
    """Check that the tofortran() method works when the character string
    contains non-ascii characters."""
    code = (
        "SELECT CASE(iflag)\n"
        "CASE(  30  ) ! This is a comment\n"
        "  IF(lwp) WRITE(*,*) ' for e1=1\xb0'\n"
        "END SELECT\n"
    )
    reader = FortranStringReader(code, ignore_comments=False)
    obj = Case_Construct(reader)
    out_str = str(obj)
    assert "for e1=1" in out_str


def test_case_construct_wrong_name(f2003_create, fake_symbol_table):
    """Check that named 'case' block has correct matching start/end name"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Case_Construct(get_reader("""\
            name: select case (n)
            case (:-1)
                a = 1
            end select wrong"""))
    assert exc_info.value.args[0].endswith("Expecting name 'name', got 'wrong'")


def test_case_construct_missing_start_name(f2003_create, fake_symbol_table):
    """Check that named 'case' block has correct matching start/end name"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Case_Construct(get_reader("""\
            select case(n)
            case (:-1)
                a = 1
            end select name"""))
    assert exc_info.value.args[0].endswith(
        "Name 'name' has no corresponding starting name"
    )


def test_case_construct_missing_end_name(f2003_create, fake_symbol_table):
    """Check that named 'case' block has correct matching start/end name"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Case_Construct(get_reader("""\
            name: select case(n)
            case (:-1)
                a = 1
            end select"""))
    assert exc_info.value.args[0].endswith("Expecting name 'name' but none given")


def test_case_construct_case_wrong_name(f2003_create, fake_symbol_table):
    """Check that named 'case' block has correct matching start/end name"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Case_Construct(get_reader("""\
            name: select case(n)
            case (:-1) wrong
                a = 1
            end select name"""))
    assert exc_info.value.args[0].endswith("Expecting name 'name', got 'wrong'")



# Copyright (c) 2017-2020 Science and Technology Facilities Council
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing tests for aspects of fparser2 related to comments"""

import pytest
from fparser.two.Fortran2003 import Program, Comment, Directive, Subroutine_Subprogram
from fparser.two.utils import walk
from fparser.api import get_reader

from fparser.two.parser import ParserFactory

# this is required to setup the fortran2003 classes
_ = ParserFactory().create(std="f2003")


def test_ignore_comments():
    """Check that the parser does throw away comments when requested"""
    tree = Program(
        get_reader(
            """\
PROGRAM a_prog
! A full line comment
PRINT *, "Hello" ! This block gets executed
END PROGRAM a_prog
    """,
            ignore_comments=True,
        )
    )
    gen = str(tree)
    assert "PROGRAM a_prog\n" '  PRINT *, "Hello"\n' "END PROGRAM a_prog" in gen
    assert "full line comment" not in gen
    assert "block gets executed" not in gen

    # Check that the default behaviour is to ignore comments
    tree = Program(get_reader("""\
PROGRAM a_prog
! A full line comment
PRINT *, "Hello" ! This block gets executed
END PROGRAM a_prog
    """))
    gen = str(tree)
    assert "full line comment" not in gen
    assert "block gets executed" not in gen


def test_simple_prog():
    """Tests simplest case of comments in a program unit"""
    tree = Program(
        get_reader(
            """\
PROGRAM a_prog
! A full line comment
PRINT *, "Hello" ! This block gets executed
END PROGRAM a_prog
    """,
            ignore_comments=False,
        )
    )
    assert (
        str(tree) == "PROGRAM a_prog\n"
        "  ! A full line comment\n"
        '  PRINT *, "Hello"\n'
        "  ! This block gets executed\n"
        "END PROGRAM a_prog\n"
    )


def test_ifthen():
    """Tests for comments within an if-then block"""
    cls = Program
    tree = cls(
        get_reader(
            """\
PROGRAM a_prog
IF(.TRUE.)THEN
! A full line comment
PRINT *, "Hello"
! Another full line comment
END IF
END PROGRAM a_prog
    """,
            ignore_comments=False,
        )
    )
    assert (
        str(tree) == "PROGRAM a_prog\n"
        "  IF (.TRUE.) THEN\n"
        "    ! A full line comment\n"
        '    PRINT *, "Hello"\n'
        "    ! Another full line comment\n"
        "  END IF\n"
        "END PROGRAM a_prog\n"
    )


def test_inline_ifthen():
    """Tests for in-line comments within an if-then block"""
    tree = Program(
        get_reader(
            """\
PROGRAM a_prog
IF(.TRUE.)THEN
PRINT *, "Hello" ! An in-line comment here
END IF
! A comment after a block
END PROGRAM a_prog
    """,
            ignore_comments=False,
        )
    )
    assert isinstance(tree, Program)
    assert (
        str(tree) == "PROGRAM a_prog\n"
        "  IF (.TRUE.) THEN\n"
        '    PRINT *, "Hello"\n'
        "    ! An in-line comment here\n"
        "  END IF\n"
        "  ! A comment after a block\n"
        "END PROGRAM a_prog\n"
    )


@pytest.mark.xfail(reason="fails to preserve formatting in fixed format")
def test_fixed_fmt():
    """Test that we handle comments in fixed-format mode"""
    reader = get_reader(
        """\
      subroutine foo
c this is a subroutine
        end subroutine foo""",
        isfree=False,
        ignore_comments=True,
    )
    cls = Subroutine_Subprogram
    obj = cls(reader)
    assert isinstance(obj, cls), repr(obj)
    assert str(obj) == "SUBROUTINE foo\nEND SUBROUTINE foo"
    assert (
        repr(obj) == "Subroutine_Subprogram(Subroutine_Stmt(None, "
        "Name('foo'), None, None), End_Subroutine_Stmt('SUBROUTINE', "
        "Name('foo')))"
    )
    reader = get_reader(
        """\
      subroutine foo
c this is a subroutine
      end subroutine foo""",
        isfree=False,
        ignore_comments=False,
    )
    cls = Subroutine_Subprogram
    obj = cls(reader)
    fort = obj.tofortran(isfix=True)
    assert isinstance(obj, cls), repr(obj)
    assert (
        fort == "      SUBROUTINE foo\nc this is a subroutine\n"
        "      END SUBROUTINE foo"
    )
    assert (
        repr(obj) == "Subroutine_Subprogram(Subroutine_Stmt(None, "
        "Name('foo'), None, None), End_Subroutine_Stmt('SUBROUTINE', "
        "Name('foo')))"
    )


@pytest.mark.xfail(reason="fails to preserve formatting in fixed format")
def test_fixed_continuation():
    """Check that we handle comments that occur within fixed-format
    continued lines"""
    reader = get_reader(
        """\
      subroutine foo
c this is a subroutine
      i = 1 +
c this comment is a problem
     &2 + 3
      end subroutine foo""",
        isfree=False,
        ignore_comments=False,
    )
    cls = Subroutine_Subprogram
    obj = cls(reader)
    assert isinstance(obj, cls)
    assert "i = 1 + 2 + 3\n     c this comment is a problem" in str(obj)


def test_prog_comments():
    """Unit tests for lines in programs containing comments"""
    cls = Program
    reader = get_reader(
        """\
   ! A troublesome comment
   program foo
     ! A full comment line
     write(*,*) my_int ! An in-line comment
    end program foo
! A really problematic comment
""",
        isfree=True,
        ignore_comments=False,
    )

    obj = cls(reader)
    assert type(obj) is Program
    # Check that the AST has the expected structure:
    # Program
    #   |--> Comment
    #   |--> Main_Program
    #   .    |--> Program_Stmt
    #   .    |--> Specification_Part
    #   .    .    \--> Implicit_Part
    #   .    .         \--> Comment
    #        |--> Execution_Part
    #        |    |--> Write_Stmt
    #        |    \--> Comment
    #   .    .
    #   .
    #   |--> Comment
    from fparser.two.Fortran2003 import Main_Program, Write_Stmt, End_Program_Stmt

    walk(obj.children, Comment, debug=True)
    assert type(obj.content[0]) is Comment
    assert str(obj.content[0]) == "! A troublesome comment"
    assert type(obj.content[1]) is Main_Program
    main_prog = obj.content[1]
    assert type(main_prog.content[1].content[0].content[0]) is Comment
    assert str(main_prog.content[1].content[0].content[0]) == "! A full comment line"
    exec_part = main_prog.content[2]
    assert type(exec_part.content[0]) is Write_Stmt
    # Check that we have the in-line comment as a second statement
    assert len(exec_part.content) == 2
    assert type(exec_part.content[1]) is Comment
    assert type(main_prog.content[3]) is End_Program_Stmt
    assert "! An in-line comment" in str(obj)
    # Check that we still have the ending comment
    assert type(obj.content[-1]) is Comment
    assert str(obj).endswith("! A really problematic comment")


def test_module_comments():
    """Tests for comments in modules"""
    source = """! This is a module
      module my_mod
        implicit none
        private
      end module my_mod
"""
    # Test when the reader is explicitly set to free-form mode
    reader = get_reader(source, isfree=True, ignore_comments=False)
    prog_unit = Program(reader)
    assert type(prog_unit.content[0]) is Comment
    assert str(prog_unit.content[0]) == "! This is a module"


def test_function_comments():
    """Tests for comments in functions"""
    source = """\
function my_mod()
  ! This is a function
  implicit none
  integer my_mod
  ! Comment1
  my_mod = 1 ! Comment2
  ! That was a function
end function my_mod
"""
    from fparser.two.Fortran2003 import Function_Subprogram

    reader = get_reader(source, isfree=True, ignore_comments=False)
    fn_unit = Function_Subprogram(reader)
    # <class 'fparser.two.Fortran2003.Function_Stmt'>
    #    <type 'NoneType'>
    #    <class 'fparser.two.Fortran2003.Name'>
    #    <type 'NoneType'>
    #    <type 'NoneType'>
    # <class 'fparser.two.Fortran2003.Specification_Part'>
    #   <class 'fparser.two.Fortran2003.Implicit_Part'>
    #     <class 'fparser.two.Fortran2003.Comment'>
    #       <type 'str'>, "'! This is a function'"
    comment = fn_unit.content[1].content[0].content[0]
    assert isinstance(comment, Comment)
    assert "! This is a function" in str(comment)
    comment = fn_unit.content[1].content[2].content[0]
    assert isinstance(comment, Comment)
    assert "! Comment1" in str(comment)
    exec_part = fn_unit.content[2]
    comment = exec_part.content[1]
    assert isinstance(comment, Comment)
    assert "! Comment2" in str(comment)
    comment = exec_part.content[2]
    assert isinstance(comment, Comment)
    assert "! That was a function" in str(comment)


def test_subroutine_comments():
    """Tests for comments in subroutines"""
    source = """\
subroutine my_mod()
  ! First comment
  implicit none
  integer my_mod
  ! Body comment
  my_mod = 1 ! Inline comment
  ! Ending comment
end subroutine my_mod
"""
    reader = get_reader(source, isfree=True, ignore_comments=False)
    fn_unit = Subroutine_Subprogram(reader)
    assert isinstance(fn_unit, Subroutine_Subprogram)
    walk(fn_unit.children, Comment, debug=True)
    spec_part = fn_unit.content[1]
    comment = spec_part.content[0].content[0]
    assert isinstance(comment, Comment)
    assert "! First comment" in str(comment)
    comment = spec_part.content[2].content[0]
    assert isinstance(comment, Comment)
    assert comment.parent is spec_part.content[2]
    assert "! Body comment" in str(comment)
    exec_part = fn_unit.content[2]
    comment = exec_part.content[1]
    assert isinstance(comment, Comment)
    assert comment.parent is exec_part
    assert "! Inline comment" in str(comment)


def test_derived_type():
    """Test for comments within declaration of derived type"""
    source = """\
type my_type ! Inline comment1
  ! First comment
  integer :: start ! Inline comment2
  integer :: stop
  ! Ending comment
end type my_type
"""
    from fparser.two.Fortran2003 import Derived_Type_Def

    reader = get_reader(source, isfree=True, ignore_comments=False)
    dtype = Derived_Type_Def(reader)
    assert isinstance(dtype, Derived_Type_Def)
    comments = [
        (1, "! Inline comment1"),
        (2, "! First comment"),
        (4, "! Inline comment2"),
        (6, "! Ending comment"),
    ]
    for comment in comments:
        idx = comment[0]
        assert isinstance(dtype.content[idx], Comment)
        assert comment[1] in str(dtype.content[idx])


def test_action_stmts():
    """Tests for comments within action statements such as allocate"""
    # We have to have our allocate() within some other block otherwise
    # we lose the comment (it is deferred until after the current,
    # continuted line is complete but once that happens, the match for
    # the allocate() is done and the reader/parser returns without
    # ever dealing with the comment).
    source = """\
if(.true.)then
allocate(my_array(size), &
         ! that's a big array
         my_array2(size))
end if
"""
    from fparser.two.Fortran2003 import If_Construct, Allocate_Stmt
    from fparser.two.utils import get_child

    reader = get_reader(source, isfree=True, ignore_comments=False)
    ifstmt = If_Construct(reader)
    assert isinstance(ifstmt, If_Construct)
    assert isinstance(ifstmt.content[1], Allocate_Stmt)
    assert "a big array" in str(ifstmt)
    cmt = get_child(ifstmt, Comment)
    assert cmt.parent is ifstmt


def test_directive_stmts():
    """Test that directives are created instead of comments when
    appropriate."""
    source = """
    Program my_prog
        integer :: x !$dir inline

        !dir$ compiler directive
        !$omp target
        !$omp loop
        do x= 1 , 100
            ! A comment!
            !!$ Another comment
        end do
    End Program"""
    reader = get_reader(
        source, isfree=True, ignore_comments=False, process_directives=True
    )
    program = Program(reader)
    out = walk(program, Directive)
    assert len(out) == 3
    assert out[0].items[0] == "!dir$ compiler directive"
    assert out[1].items[0] == "!$omp target"
    assert out[2].items[0] == "!$omp loop"

    assert out[2].tostr() == "!$omp loop"

    # Check the restore_reader works correctly for directive.
    old = reader.get_item()
    assert old is None
    out[2].restore_reader(reader)
    old = reader.get_item()
    assert old is not None

    out = walk(program, Comment)
    comments = 0
    for comment in out:
        if comment.items[0] != "":
            comments = comments + 1
    assert comments == 3
    assert str(out[1]) == "!$dir inline"
    assert str(out[3]) == "! A comment!"
    assert str(out[4]) == "!!$ Another comment"

    # Check that passing something that isn't a comment into a Directive
    # __new__ call doesn't create a Directive.
    out = Directive(program)
    assert out is None

    # Fixed form test that directives are handled.
    reader = get_reader(
        """\
      program foo
cdir$ This is a directive
C     This is a comment
C$    integer omp_get_thread_num
        end program foo""",
        isfree=False,
        ignore_comments=False,
        process_directives=True,
    )
    program = Program(reader)
    out = walk(program, Directive)
    assert len(out) == 1
    assert out[0].items[0] == "cdir$ This is a directive"


@pytest.mark.parametrize(
    "directive,expected,free",
    [
        ("!$dir always", ("!$dir always",), True),
        ("!dir$ always", ("!dir$ always",), True),
        ("!gcc$ vector", ("!gcc$ vector",), True),
        ("!$acc loop", ("!$acc loop",), True),
        ("!$omp parallel", ("!$omp parallel",), True),
        ("!$ompx parallel", ("!$ompx parallel",), True),
        ("c$omp parallel", ("c$omp parallel",), False),
        ("c$omx parallel", ("c$omx parallel",), False),
        ("!$omx parallel", ("!$omx parallel",), False),
        ("*$omp parallel", ("*$omp parallel",), False),
        ("c$omx parallel", ("c$omx parallel",), False),
        ("*$omx parallel", ("*$omx parallel",), False),
        ("!$DIR ALWAYS", ("!$DIR ALWAYS",), True),
        ("c$OMX PARALLEL", ("c$OMX PARALLEL",), False),
        ("!$omp parallel&\n!$omp&do", ("!$omp parallel&", "!$omp&do"), True),
        (
            "c$omp parallel do\nc$omp+shared(a,b,c)",
            ("c$omp parallel do", "c$omp+shared(a,b,c)"),
            False,
        ),
        ("!!$omp parallel", (), True),
    ],
)
def test_all_directive_formats(directive, expected, free):
    """Parameterized test to ensure that all directive formats are
    correctly recognized."""
    # Generate the source code
    if free:
        source = """Program my_prog
            integer :: x
        """
        source = source + directive + "\n"
        source = source + """          do x= 1 , 100
            end do
        End Program"""
    else:
        source = """\
      program foo
"""
        source = source + directive + "\n"
        source = source + "        end program foo"

    reader = get_reader(
        source, isfree=free, ignore_comments=False, process_directives=True
    )
    program = Program(reader)
    out = walk(program, Directive)
    assert len(out) == len(expected)
    for i, direc in enumerate(out):
        assert direc.items[0] == expected[i]

    # Test that we correctly get directives without ignore_comments=False.
    reader = get_reader(source, isfree=free, process_directives=True)
    program = Program(reader)
    out = walk(program, Directive)
    assert len(out) == len(expected)
    for i, direc in enumerate(out):
        assert direc.items[0] == expected[i]


@pytest.mark.parametrize(
    "directive,expected,free",
    [
        ("!$dir always", ("!$dir always",), True),
        ("!dir$ always", ("!dir$ always",), True),
        ("!gcc$ vector", ("!gcc$ vector",), True),
        ("!$omp parallel", ("!$omp parallel",), True),
        ("!$ompx parallel", ("!$ompx parallel",), True),
        ("c$omp parallel", ("c$omp parallel",), False),
        ("c$omx parallel", ("c$omx parallel",), False),
        ("!$omx parallel", ("!$omx parallel",), False),
        ("*$omp parallel", ("*$omp parallel",), False),
        ("c$omx parallel", ("c$omx parallel",), False),
        ("*$omx parallel", ("*$omx parallel",), False),
        ("!$DIR ALWAYS", ("!$DIR ALWAYS",), True),
        ("c$OMX PARALLEL", ("c$OMX PARALLEL",), False),
        ("!$omp parallel&\n!$omp&do", ("!$omp parallel&", "!$omp&do"), True),
        (
            "c$omp parallel do\nc$omp+shared(a,b,c)",
            ("c$omp parallel do", "c$omp+shared(a,b,c)"),
            False,
        ),
        ("!!$omp parallel", ("!!$omp parallel",), True),
    ],
)
def test_directives_as_comments(directive, expected, free):
    """Parameterized test to ensure all directives produce comments when
    process_directives is disabled."""
    # Generate the source code
    if free:
        source = """Program my_prog
            integer :: x
        """
        source = source + directive + "\n"
        source = source + """          do x= 1 , 100
            end do
        End Program"""
    else:
        source = """\
      program foo
"""
        source = source + directive + "\n"
        source = source + "        end program foo"
    # Test that we get the expected comments with comments only
    reader = get_reader(
        source, isfree=free, ignore_comments=False, process_directives=False
    )
    program = Program(reader)
    out = walk(program, Comment)
    # Check that we have the correct number of comments.
    assert len(out) == len(expected)
    # Check that the comments contain the correct strings.
    for i, direc in enumerate(out):
        assert direc.items[0] == expected[i]


def test_inline_directive_is_comment():
    """Inline comments on statements should be comments even if containing
    directive markers."""
    source = """Program my_prog
    integer :: x !$dir directive

    x = 1 !$dir comment
    end program
    """
    reader = get_reader(source, ignore_comments=False, process_directives=True)
    program = Program(reader)
    out = walk(program, Directive)
    assert len(out) == 0



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 Cray-pointers: This file tests the support for a
Cray-pointee array specification.

"""

import pytest
from fparser.two.Fortran2003 import Cray_Pointee_Array_Spec
from fparser.two.utils import NoMatchError


def test_cray_pointee_array_spec(f2003_create):
    """Check that Cray-pointee array specifications are parsed correctly."""
    for myinput in ["n", "0 : n", "n, m", "5, *", "*", "0 : 1, 2 : *"]:
        ast = Cray_Pointee_Array_Spec(myinput)
        assert myinput in str(ast)


def test_errors(f2003_create):
    """Check that syntax errors produce a NoMatchError exception."""
    for myinput in ["", "  ", "2n", "2 n", "*, n"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Cray_Pointee_Array_Spec(myinput)
        assert "Cray_Pointee_Array_Spec: '{0}'".format(myinput) in str(excinfo.value)


def test_unsupported(f2003_create):
    """Check that unsupported specifications produce a
    NoMatchError exception.

    """
    for myinput in [":", "2,:"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Cray_Pointee_Array_Spec(myinput)
        assert "Cray_Pointee_Array_Spec: '{0}'".format(myinput) in str(excinfo.value)



# Copyright (c) 2023 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R823

type-guard-stmt is TYPE IS ( type-spec ) [ select-construct-name ]
                or CLASS IS ( type-spec ) [ select-construct-name ]
                or CLASS DEFAULT [ select-construct-name ]

"""

import pytest

from fparser.two.Fortran2003 import Type_Guard_Stmt
from fparser.two.utils import NoMatchError


@pytest.mark.usefixtures("f2003_create")
def test_type_is():
    """Test that the match works as expected when trying to match 'type
    is'

    """
    # Space before 'type is'
    tcls = Type_Guard_Stmt
    obj = tcls("  type is (real)")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "TYPE IS (REAL)"

    # Mixed case keyword
    obj = tcls("TyPe Is (mytype)")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "TYPE IS (mytype)"

    # Invalid 'type' keyword
    with pytest.raises(NoMatchError) as info:
        _ = tcls("TYP Is (mytype)")
    assert "Type_Guard_Stmt: 'TYP Is (mytype)'" in str(info.value)

    # Invalid 'is' keyword
    with pytest.raises(NoMatchError) as info:
        _ = tcls("TYPE IZ (mytype)")
    assert "Type_Guard_Stmt: 'TYPE IZ (mytype)'" in str(info.value)


@pytest.mark.usefixtures("f2003_create")
def test_class_is():
    """Test that the match works as expected when trying to match 'class
    is'

    """
    # Space before 'class is'
    tcls = Type_Guard_Stmt
    obj = tcls("  class is (real)")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "CLASS IS (REAL)"

    # Mixed case keyword
    obj = tcls("ClAsS Is (mytype)")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "CLASS IS (mytype)"

    # Invalid 'class' keyword
    with pytest.raises(NoMatchError) as info:
        _ = tcls("CLAS IS (mytype)")
    assert "Type_Guard_Stmt: 'CLAS IS (mytype)'" in str(info.value)

    # Invalid 'is' keyword
    with pytest.raises(NoMatchError) as info:
        _ = tcls("CLASS IZ (mytype)")
    assert "Type_Guard_Stmt: 'CLASS IZ (mytype)'" in str(info.value)


@pytest.mark.usefixtures("f2003_create")
def test_default():
    """Test that the match works as expected when trying to match 'class
    default'

    """
    # Space before 'class default'
    tcls = Type_Guard_Stmt
    obj = tcls("  class default")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "CLASS DEFAULT"

    # Mixed case keyword
    obj = tcls("ClAsS DeFaUlT")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "CLASS DEFAULT"

    # Invalid 'class' keyword
    with pytest.raises(NoMatchError) as info:
        _ = tcls("CLAS DEFAULT")
    assert "Type_Guard_Stmt: 'CLAS DEFAULT'" in str(info.value)

    # Invalid 'default' keyword
    with pytest.raises(NoMatchError) as info:
        _ = tcls("CLASS DERFAULT")
    assert "Type_Guard_Stmt: 'CLASS DERFAULT'" in str(info.value)

    # Invalid optional name after CLASS DEFAULT
    with pytest.raises(NoMatchError) as info:
        _ = tcls("CLASS DEFAULT (mytype)")
    assert "Type_Guard_Stmt: 'CLASS DEFAULT (mytype)'" in str(info.value)

    # Valid name after CLASS DEFAULT
    obj = tcls("class default myname")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "CLASS DEFAULT myname"


@pytest.mark.usefixtures("f2003_create")
def test_invalid_keyword():
    """Test that the match fails if the initial matching keyword is not
    CLASS or TYPE.

    """
    tcls = Type_Guard_Stmt
    with pytest.raises(NoMatchError) as info:
        _ = tcls("invalid")
    assert "Type_Guard_Stmt: 'invalid'" in str(info.value)


@pytest.mark.usefixtures("f2003_create")
def test_clause():
    """Test matching works as expected when trying to match 'class is
    (mytype)'. Also covers case with 'type is' so no need for
    additional tests.

    """
    # invalid ( location
    tcls = Type_Guard_Stmt
    with pytest.raises(NoMatchError) as info:
        _ = tcls(" class is m(ytype)")
    assert "Type_Guard_Stmt: ' class is m(ytype)'" in str(info.value)

    # non-existant )
    with pytest.raises(NoMatchError) as info:
        _ = tcls(" class is (mytype")
    assert "Type_Guard_Stmt: ' class is (mytype'" in str(info.value)

    # no content in ()
    with pytest.raises(NoMatchError) as info:
        _ = tcls(" class is ()")
    assert "Type_Guard_Stmt: ' class is ()'" in str(info.value)

    # valid string
    obj = tcls(" class is (mytype)")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "CLASS IS (mytype)"

    # valid string with optional name
    obj = tcls(" class is (mytype) name")
    assert isinstance(obj, tcls), repr(obj)
    assert str(obj) == "CLASS IS (mytype) name"


@pytest.mark.usefixtures("f2003_create")
def test_tostr():
    """Test the tostr() method works as expected"""

    tcls = Type_Guard_Stmt
    obj = tcls("class default")
    assert obj.tostr() == "CLASS DEFAULT"

    obj = tcls("class default name")
    assert obj.tostr() == "CLASS DEFAULT name"

    obj = tcls("class is (real)")
    assert obj.tostr() == "CLASS IS (REAL)"

    obj = tcls("class is (real) name")
    assert obj.tostr() == "CLASS IS (REAL) name"

    obj = tcls("type is (real)")
    assert obj.tostr() == "TYPE IS (REAL)"

    obj = tcls("type is (real) name")
    assert obj.tostr() == "TYPE IS (REAL) name"


@pytest.mark.usefixtures("f2003_create")
def test_get_end_name():
    """Test the get_end_name method works as expected."""

    tcls = Type_Guard_Stmt
    obj = tcls("type is (real)")
    assert obj.get_end_name() is None

    obj = tcls("type is (real) name")
    assert obj.get_end_name() == "name"



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 End_Submodule_Stmt rule R1119
"""

from fparser.two.utils import EndStmtBase


class End_Submodule_Stmt(EndStmtBase):  # R1119
    """
    Fortran 2008 rule R1119
    end-submodule-stmt is END [ SUBMODULE [ submodule-name ] ]

    """

    subclass_names = []
    use_names = ["Submodule_Name"]

    @staticmethod
    def match(fstring):
        """Check whether the input matches the rule

        param string fstring : contains the Fortran that we are trying
        to match

        :return: instances of the Classes that have matched if there
        is a match or `None` if there is no match

        """
        # Avoid circular dependencies by importing here.
        # pylint: disable=import-outside-toplevel
        from fparser.two.Fortran2008 import Submodule_Name

        return EndStmtBase.match("SUBMODULE", Submodule_Name, fstring)

    def get_name(self):  # C1114
        """Fortran 2008 constraint C1114 return the submodule name as
        specified by the end submodule statement or `None` if one is
        not specified. This is used by the base class to check whether
        this name matches the submodule name.

        :return: the name of the submodule stored in a Name class
        :return type: :py:class:`fparser.two.Fortran2003.Name` or `None`

        """
        return self.items[1]



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------
"""This module provides the Fortran2008-specific version of a partial
implementation of the block-do-construct rule r814. fparser splits
this rule into a label and nonlabel version.

"""

from fparser.two.Fortran2003 import (
    Block_Nonlabel_Do_Construct as Block_Nonlabel_Do_Construct_2003,
)
from fparser.two.Fortran2008.nonlabel_do_stmt_r817 import Nonlabel_Do_Stmt


class Block_Nonlabel_Do_Construct(Block_Nonlabel_Do_Construct_2003):
    """Subclass the 2003 version so that this class will import the
    Fortran2008 Nonlabel_Do_Stmt class

    """

    @staticmethod
    def nonlabel_do_stmt_cls():
        """
        :returns: Fortran2008 Nonlabel_Do_Stmt class.
        :rtype: :py:class:`fparser.two.Fortran2008.Nonlabel_Do_Stmt`

        """
        return Nonlabel_Do_Stmt



# Copyright (c) 2023 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2008 rule R818

loop-control is [ , ] do-variable = scalar-int-expr , scalar-int-expr
                   [ , scalar-int-expr ]
                or [ , ] WHILE ( scalar-logical-expr )
                or [ , ] CONCURRENT forall-header

Extends the Fortran2003 rule R830 with the additional CONCURRENT clause.

"""

import pytest
from fparser.two.Fortran2008 import Loop_Control
from fparser.two.utils import NoMatchError


def test_f2003_match():
    """Test that there is a match if the string contains a F2003 form of
    the rule and that the F2003 tostr() is as expected.

    """
    result = Loop_Control("while (.true.)")
    assert isinstance(result, Loop_Control)
    assert str(result) == "WHILE (.TRUE.)"


def test_start_space():
    """Test that there is a match if the string contains white space at
    the start and end and that the tostr() output is as expected.

    """
    result = Loop_Control(" concurrent (i=1:10) ")
    assert isinstance(result, Loop_Control)
    assert str(result) == "CONCURRENT (i = 1 : 10)"


def test_delim():
    """Test that there is a match if the string contains an options
    delimiter at the start and that the tostr() output is as expected.

    """
    result = Loop_Control(" , concurrent (i=1:10)")
    assert isinstance(result, Loop_Control)
    assert str(result) == ", CONCURRENT (i = 1 : 10)"


@pytest.mark.parametrize(
    "string",
    [
        "",
        ": concurrent (i=1:10)",
        "concurren (i=1:10)",
        "concurrent",
        "concurrent invalid",
    ],
)
def test_invalid(string):
    """Test that there is no match for various invalid input strings."""
    with pytest.raises(NoMatchError):
        _ = Loop_Control(string)



# Copyright (c) 2018 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Test Fortran 2003 rule R451 : This file tests the support for specific
binding and passed-object dummy argument within the type-bound procedure
part of a derived type.
"""

import pytest
from fparser.two.utils import NoMatchError, InternalError
from fparser.two.Fortran2003 import Specific_Binding


def test_valid(f2003_create):
    """Test that valid Specific_Binding statements are parsed
    correctly.

    """

    # simple statement
    obj = Specific_Binding("procedure sub")
    assert str(obj) == "PROCEDURE sub"

    # simple statement spaces
    obj = Specific_Binding("  procedure  sub  ")
    assert str(obj) == "PROCEDURE sub"

    # multi-case procedure keyword
    obj = Specific_Binding("PrOcEdUrE sub")
    assert str(obj) == "PROCEDURE sub"

    # simple statement with ::
    obj = Specific_Binding("procedure :: sub")
    assert str(obj) == "PROCEDURE :: sub"

    # simple statement with :: no space
    obj = Specific_Binding("procedure::sub")
    assert str(obj) == "PROCEDURE :: sub"

    # statement with procedure name
    obj = Specific_Binding("procedure :: sub => boat")
    assert str(obj) == "PROCEDURE :: sub => boat"

    # statement with procedure name no space
    obj = Specific_Binding("procedure::sub=>boat")
    assert str(obj) == "PROCEDURE :: sub => boat"

    # statement with interface name
    obj = Specific_Binding("procedure ( interface ) sub")
    assert str(obj) == "PROCEDURE(interface) sub"

    # statement with interface name no space
    obj = Specific_Binding("procedure(interface)::sub")
    assert str(obj) == "PROCEDURE(interface) :: sub"

    # statement with binding list
    obj = Specific_Binding("procedure , PASS :: sub")
    assert str(obj) == "PROCEDURE, PASS :: sub"

    # statement with binding list no space
    obj = Specific_Binding("procedure,PASS::sub")
    assert str(obj) == "PROCEDURE, PASS :: sub"

    # statement with multi-binding list
    obj = Specific_Binding("procedure , PASS , PUBLIC, deferred :: sub")
    assert str(obj) == "PROCEDURE, PASS, PUBLIC, DEFERRED :: sub"

    # statement with multi-binding list no space
    obj = Specific_Binding("procedure,PASS,PUBLIC,DEFERRED::sub")
    assert str(obj) == "PROCEDURE, PASS, PUBLIC, DEFERRED :: sub"

    # statement with interface name and binding list
    obj = Specific_Binding("procedure ( interface ) , PASS :: sub")
    assert str(obj) == "PROCEDURE(interface), PASS :: sub"

    # statement with interface name and binding list no space
    obj = Specific_Binding("procedure(interface),PASS::sub")
    assert str(obj) == "PROCEDURE(interface), PASS :: sub"

    # statement with binding list and procedure name
    obj = Specific_Binding("procedure , PASS :: sub => boat")
    assert str(obj) == "PROCEDURE, PASS :: sub => boat"

    # statement with binding list and procedure name no space
    obj = Specific_Binding("procedure,PASS::sub=>boat")
    assert str(obj) == "PROCEDURE, PASS :: sub => boat"


def test_invalid():
    """Test that parsing invalid Fortran syntax for
    Specific_Binding statements raises an appropriate error."""

    # includes tests for breaking C456 and C457
    for string in [
        "",
        "  ",
        "procedure  ",
        "procedure ::  ",
        "procedur sub",
        "proceduresub",
        "procedure : sub",
        "procedure sub => boat",  # C456
        "procedure :: sub =>",
        "procedure (interface sub",
        "procedure interface) sub",
        "procedure (interface) :: sub => boat",  # C457
        "procedure pass sub",
        "procedure, pass sub",
        "procedure, pass public sub",
        "procedure pass, public sub",
        "procedure pass :: sub",
        "procedure sub sub2",
        "procedure :: sub = boat",
        "procedure :: sub > boat",
        "procedure :: sub => boat boat2",
    ]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Specific_Binding(string)
        assert "Specific_Binding: '{0}'".format(string) in str(excinfo.value)


def test_tostr_invalid(f2003_create, monkeypatch):
    """Test that invalid input raises an exception"""

    # test internal error in tostr() when the items list is not the
    # expected size
    obj = Specific_Binding("procedure hello")
    monkeypatch.setattr(obj, "items", ["procedure"])
    with pytest.raises(InternalError) as excinfo:
        _ = str(obj)
    assert "tostr() has '1' items, but expecting 5" in str(excinfo.value)



# -----------------------------------------------------------------------------
# Copyright (c) 2022-2023 Science and Technology Facilities Council.
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""Module containing tests for the ModuleUse class."""

import pytest
from fparser.two.symbol_table import ModuleUse


def test_moduse_constructor():
    """Checks for the ModuleUse constructor."""
    # Name only.
    muse1 = ModuleUse("Nancy")
    assert muse1.name == "nancy"
    assert muse1.only_list is None
    assert muse1.rename_list is None
    assert muse1.wildcard_import is True
    # pylint: disable=use-implicit-booleaness-not-comparison
    assert muse1.symbol_names == []
    # Only-list.
    muse2 = ModuleUse("peggy", only_list=[("amazon", None), ("swallow", None)])
    assert sorted(muse2.only_list) == ["amazon", "swallow"]
    assert muse2.wildcard_import is False
    assert muse2.rename_list is None
    assert muse2.symbol_names == ["amazon", "swallow"]
    # Only-list with renaming.
    muse3 = ModuleUse("roger", only_list=[("swallow", None), ("amazon", "dingy")])
    assert sorted(muse3.only_list) == ["amazon", "swallow"]
    assert muse3.get_declared_name("amazon") == "dingy"
    assert muse3.symbol_names == ["swallow", "amazon"]
    # Rename list.
    muse4 = ModuleUse("susan", rename_list=[("swallow", "amazon"), ("boat", "dingy")])
    assert sorted(muse4.rename_list) == ["boat", "swallow"]
    assert muse4.get_declared_name("swallow") == "amazon"
    assert muse4.get_declared_name("boat") == "dingy"
    assert muse4.symbol_names == ["swallow", "boat"]


def test_moduse_constructor_validate():
    """Checks for the type-checking performed by the constructor."""
    # Module name.
    with pytest.raises(TypeError) as err:
        _ = ModuleUse(1)
    assert "name of the module must be a str but got" in str(err.value)
    # Check that it calls _validate_tuple_list.
    with pytest.raises(TypeError) as err:
        _ = ModuleUse("titty", only_list="polly")
    assert "If present, the only_list must be a list but got 'str'" in str(err.value)
    with pytest.raises(TypeError) as err:
        _ = ModuleUse("titty", rename_list="polly")
    assert "If present, the rename_list must be a list but got 'str'" in str(err.value)
    with pytest.raises(TypeError) as err:
        _ = ModuleUse("titty", only_list=[("polly", 1)])
    assert (
        "If present, the only_list must be a list of 2-tuples of "
        "(str, str | NoneType) but got: [('polly', 1)]" in str(err.value)
    )
    with pytest.raises(TypeError) as err:
        _ = ModuleUse("titty", rename_list=[("polly", 1)])
    assert (
        "If present, the rename_list must be a list of 2-tuples of "
        "(str, str) but got: [('polly', 1)]" in str(err.value)
    )


def test_moduse_validate_tuple_list():
    """Checks for the _validate_tuple_list() method."""
    with pytest.raises(TypeError) as err:
        _ = ModuleUse._validate_tuple_list("only", ["polly"])
    assert (
        "If present, the only_list must be a list of 2-tuples but "
        "got: ['polly']" in str(err.value)
    )
    with pytest.raises(TypeError) as err:
        _ = ModuleUse._validate_tuple_list("only", [("polly", "gibber", 1)])
    assert (
        "If present, the only_list must be a list of 2-tuples but "
        "got: [('polly', 'gibber', 1)]" in str(err.value)
    )


def test_moduse_update():
    """Checks for the update() method."""
    moduse = ModuleUse("flint")
    assert moduse.wildcard_import is True
    # Wrong argument type.
    with pytest.raises(TypeError) as err:
        moduse.update("hello")
    assert (
        "update() must be supplied with an instance of ModuleUse but got "
        "'str'" in str(err.value)
    )
    # Wrong module name.
    moduse1 = ModuleUse("houseboat", only_list=[("cannon", None)])
    with pytest.raises(ValueError) as err:
        moduse.update(moduse1)
    assert (
        "ModuleUse supplied to update() is for module 'houseboat' but this "
        "ModuleUse is for module 'flint'" in str(err.value)
    )
    # Add an import with an ONLY: clause.
    moduse2 = ModuleUse("flint", only_list=[("cannon", None)])
    moduse.update(moduse2)
    assert moduse.wildcard_import is True
    assert moduse.only_list == ["cannon"]
    assert moduse.rename_list is None
    # Add an import with a Rename_List.
    moduse3 = ModuleUse("Flint", rename_list=[("uncle", "jim")])
    moduse.update(moduse3)
    assert moduse.only_list == ["cannon"]
    assert moduse.rename_list == ["uncle"]
    assert moduse.get_declared_name("uncle") == "jim"
    assert moduse.symbol_names == ["cannon", "uncle"]
    # Add further Only and Rename lists.
    moduse4 = ModuleUse("Flint", rename_list=[("rogEr", "giBber")])
    moduse.update(moduse4)
    assert moduse.get_declared_name("uncle") == "jim"
    assert moduse.get_declared_name("roger") == "gibber"
    assert moduse.symbol_names == ["cannon", "uncle", "roger"]
    assert sorted(moduse.rename_list) == ["roger", "uncle"]
    moduse5 = ModuleUse("Flint", only_list=[("maTe", "peggy")])
    moduse.update(moduse5)
    assert sorted(moduse.only_list) == ["cannon", "mate"]
    assert moduse.get_declared_name("roger") == "gibber"
    assert moduse.get_declared_name("mate") == "peggy"
    assert moduse.symbol_names == ["cannon", "uncle", "roger", "mate"]
    # Adding a wildcard import.
    moduse6 = ModuleUse("flInt")
    moduse5.update(moduse6)
    assert moduse5.wildcard_import is True


def test_moduse_lookup():
    """Tests for the lookup() method."""
    moduse = ModuleUse("flint")
    with pytest.raises(KeyError):
        moduse.lookup("polly")
    moduse2 = ModuleUse("flint", only_list=[("mate", None), ("cannON", None)])
    sym = moduse2.lookup("caNNon")
    assert sym.name == "cannon"
    # As this symbol is imported, we don't know its type.
    assert sym.primitive_type == "unknown"



# Copyright (c) 2022-2023 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing pytest tests for the support of the Fortran2008
Block construct."""

import re
import pytest

from fparser.api import get_reader
from fparser.two.Fortran2008 import Block_Construct, Block_Stmt
from fparser.two.symbol_table import SYMBOL_TABLES
from fparser.two.utils import FortranSyntaxError, ScopingRegionMixin, walk


def test_block():
    """Test that the Block_Construct matches as expected."""
    block = Block_Construct(get_reader("""\
            block
               integer :: b = 4
               a = 1 + b
            end block
            """))
    assert isinstance(block.children[0], Block_Stmt)
    assert isinstance(block.children[0], ScopingRegionMixin)
    name = block.children[0].get_scope_name()
    assert re.match(r"block:[\d+]", name)
    assert "BLOCK\n  INTEGER :: b = 4\n  a = 1 + b\nEND BLOCK" in str(block)


@pytest.mark.parametrize(
    "before, after", [("", ""), ("b = 2.0 * b", ""), ("", "b = 2.0 * b")]
)
def test_block_new_scope(f2008_parser, before, after):
    """Test that a Block_Construct creates a new scoping region."""
    block = f2008_parser(get_reader(f"""\
            program foo
            integer :: b = 3
            {before}
            block
               integer :: b = 4
               a = 1 + b
            end block
            {after}
            end program foo
            """))

    assert "BLOCK\nINTEGER :: b = 4\na = 1 + b\nEND BLOCK" in str(block).replace(
        "  ", ""
    )
    tables = SYMBOL_TABLES
    assert list(tables._symbol_tables.keys()) == ["foo"]
    table = SYMBOL_TABLES.lookup("foo")
    assert len(table.children) == 1
    assert re.match(r"block:[\d+]", table.children[0].name)


def test_block_in_if(f2008_parser):
    """Test that a Block may appear inside an IF."""
    ptree = f2008_parser(get_reader("""\
            program foo
            integer :: b = 3
            if (b == 2) then
              block
                real :: tmp
                tmp = ATAN(0.5)
                b = NINT(tmp)
              end block
            end if
            end program foo
            """))
    blocks = walk([ptree], Block_Construct)
    assert len(blocks) == 1


def test_named_block():
    """
    Test that a named block construct is correctly captured and also
    reproduced.

    """
    block = Block_Construct(get_reader("""\
            foo: block
               integer :: b = 4
               a = 1 + b
            end block foo
            """))

    assert "foo:BLOCK\n  INTEGER :: b = 4\n  a = 1 + b\nEND BLOCK foo" in str(block)


def test_end_block_missing_start_name():  # C808
    """
    Test Constraint 808 - that a name on the 'end block' must correspond
    with the same name on the 'block'.

    """
    with pytest.raises(FortranSyntaxError) as err:
        Block_Construct(get_reader("""\
                block
                end block foo
                """))
    assert "Name 'foo' has no corresponding starting name" in str(err)


def test_end_block_missing_end_name():  # C808
    """
    Test that a named block that is missing a name on its 'end block' statement
    results in a syntax error.

    """
    with pytest.raises(FortranSyntaxError) as err:
        Block_Construct(get_reader("""\
                foo: block
                end block
                """))
    assert "Expecting name 'foo' but none given" in str(err)


def test_end_block_wrong_name():  # C808
    """
    Test that an incorrect name on the end block statement results in a
    syntax error.

    """
    with pytest.raises(FortranSyntaxError) as err:
        Block_Construct(get_reader("""\
                foo: block
                end block bar
                """))
    assert "Expecting name 'foo', got 'bar'" in str(err)


def test_block_in_subroutine(f2008_parser):
    """
    Check that we get two, nested symbol tables when a routine contains
    a Block construct.

    """
    code = """\
            program my_prog
            real :: a
            a = -1.0
            if (a < 0.0) then
             rocking: block
               real :: b
               b = 42.0
               a = b
             end block rocking
            else
             block
               real :: c
               c = 42.0 / 5.0
               a = 10.0 * c
             end block
            end if
            end program my_prog
            """
    _ = f2008_parser(get_reader(code))
    tables = SYMBOL_TABLES
    assert list(tables._symbol_tables.keys()) == ["my_prog"]
    table = SYMBOL_TABLES.lookup("my_prog")
    assert len(table.children) == 2
    assert table.children[0].name == "rocking"
    assert re.match(r"block:[\d+]", table.children[1].name)
